local ADDON, Package = ...

local Kernel = {}
local String = {}
local Enum = {}
local Utils = {}
local Observable = {
  observers = {}
}
local Widget = {}

--------------------------------------------------------------------------------
-- Kernel
--------------------------------------------------------------------------------
function Kernel.is_table(any) return type(any) == "table" end

function Kernel.is_number(arg) return type(arg) == "number" end

function Kernel.is_string(arg) return type(arg) == "string" end

function Kernel.noop() end

function Kernel.dump(...)
  if not IsAddOnLoaded("Blizzard_DebugTools") then
    UIParentLoadAddOn("Blizzard_DebugTools")
  end

  for at = 1, select("#", ...) do
    DevTools_Dump(select(at, ...))
  end
end

--------------------------------------------------------------------------------
-- String
--------------------------------------------------------------------------------
function String.camelcase(value)
  return string.gsub(" " .. value, "%W%l", string.upper):gsub("[_ ]", "")
end

--------------------------------------------------------------------------------
-- Enum
--------------------------------------------------------------------------------
function Enum.clone(enumerable)
  local stash = {}

  for k, v in pairs(enumerable) do
    if Kernel.is_table(v) then
      stash[k] = Enum.clone(v)
    else
      stash[k] = v
    end
  end

  return stash
end

function Enum.merge(to, from, override)
  local stash = Enum.clone(to)

  for k, v in pairs(from) do
    if Kernel.is_table(v) and Kernel.is_table(stash[k]) then
      stash[k] = Enum.merge(stash[k], v, override)
    else
      if stash[k] == nil or override ~= false then
        stash[k] = v
      end
    end
  end

  return stash
end

function Enum.omit(enumerable, keys, depth)
  local stash = {}
  local depth = depth or 0

  for k, v in pairs(enumerable) do
    if not Enum.is_member(keys, k) then
      if Kernel.is_table(v) and depth > 0 then
        stash[k] = Enum.omit(v, keys, depth - 1)
      else
        stash[k] = v
      end
    end
  end

  return stash
end

function Enum.rename(enumerable, from, to)
  local stash = {}

  for k, v in pairs(enumerable) do
    local i = k == from and to or k

    if Kernel.is_table(v) then
      stash[i] = Enum.rename(v, from, to)
    else
      stash[i] = v
    end
  end

  return stash
end

function Enum.filter(enumerable, value)
  local stash = {}

  for _, entry in pairs(enumerable) do
    if entry ~= value then
      table.insert(stash, entry)
    end
  end

  return stash
end

function Enum.is_member(enumerable, value)
  for _, entry in pairs(enumerable) do
    if entry == value then
      return true
    end
  end

  return false
end

--------------------------------------------------------------------------------
-- Utils
--------------------------------------------------------------------------------
local RACE_ICON_NAMES = {
  HUMAN = {"raceicon-human-male", "raceicon-human-female"},
  DWARF = {"raceicon-dwarf-male", "raceicon-dwarf-female"},
  GNOME = {"raceicon-gnome-male", "raceicon-gnome-female"},
  NIGHTELF = {"raceicon-nightelf-male", "raceicon-nightelf-female"},
  BLOODELF = {"raceicon-bloodelf-male", "raceicon-bloodelf-female"},
  DRAENEI = {"raceicon-draenei-male", "raceicon-draenei-female"},
  TAUREN = {"raceicon-tauren-male", "raceicon-tauren-female"},
  SCOURGE = {"raceicon-undead-male", "raceicon-undead-female"},
  TROLL = {"raceicon-troll-male", "raceicon-troll-female"},
  ORC = {"raceicon-orc-male", "raceicon-orc-female"}
}

function Utils.create_character_name(attrs)
  return RAID_CLASS_COLORS[attrs.class]:WrapTextInColorCode(attrs.name)
end

function Utils.create_character_icon(attrs)
  return CreateAtlasMarkup(RACE_ICON_NAMES[attrs.race][attrs.sex], 16, 16)
end

function Utils.create_character_information()
  local UNIT = "player"

  return {
    realm = GetRealmName(),
    faction = UnitFactionGroup(UNIT):upper(),
    sex = UnitSex(UNIT) - 1,
    race = select(2, UnitRace(UNIT)):upper(),
    name = UnitName(UNIT),
    class = select(2, UnitClass(UNIT)),
    level = UnitLevel(UNIT)
  }
end

function Utils.purge_keyboard_focus()
  local frame = GetCurrentKeyBoardFocus()
  if frame and frame.ClearFocus then
    frame:ClearFocus()
  end
end

--------------------------------------------------------------------------------
-- Observable
--------------------------------------------------------------------------------
function Observable.broadcast(scope, ...)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  for at = 1, #Observable.observers[scope] do
    Observable.observers[scope][at](...)
  end
end

function Observable.subscribe(scope, observer)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  if not Enum.is_member(Observable.observers[scope], observer) then
    table.insert(Observable.observers[scope], observer)
  end
end

function Observable.unsubscribe(scope, observer)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  Observable.observers[scope] = Enum.filter(Observable.observers[scope], observer)
end

--------------------------------------------------------------------------------
-- Widget
--------------------------------------------------------------------------------
function Widget.create_window(optional_attrs, parent)
  local default_attrs = {
    size = {128, 128},
    name = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", attrs.name, parent)
  frame:Hide()
  frame:EnableKeyboard(true)
  frame:EnableMouse(true)
  frame:EnableMouseWheel(true)
  frame:SetToplevel(true)
  frame:SetMovable(true)
  frame:SetFrameStrata("DIALOG")
  frame:SetSize(unpack(attrs.size))

  local background = frame:CreateTexture(nil, "BACKGROUND")
  background:SetAllPoints(frame)
  background:SetTexture(131071)

  local close_button_border = frame:CreateTexture(nil, "BORDER")
  close_button_border:SetTexture(131073)
  close_button_border:SetSize(32, 32)
  close_button_border:SetPoint("TOPRIGHT", 2, 2)

  local close_button = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
  close_button:SetPoint("TOPRIGHT", 6, 6)

  local border_topleft = frame:CreateTexture(nil, "ARTWORK")
  border_topleft:SetAtlas("battlefieldminimap-border-topleft", true)
  border_topleft:SetPoint("TOPLEFT", -10, 10)

  local border_topright = frame:CreateTexture(nil, "ARTWORK")
  border_topright:SetAtlas("battlefieldminimap-border-topright", true)
  border_topright:SetPoint("TOPRIGHT", 8, 10)

  local border_bottomleft = frame:CreateTexture(nil, "ARTWORK")
  border_bottomleft:SetAtlas("battlefieldminimap-border-bottomleft", true)
  border_bottomleft:SetPoint("BOTTOMLEFT", -10, -8)

  local border_bottomright = frame:CreateTexture(nil, "ARTWORK")
  border_bottomright:SetAtlas("battlefieldminimap-border-bottomright", true)
  border_bottomright:SetPoint("BOTTOMRIGHT", 8, -8)

  local border_top = frame:CreateTexture(nil, "ARTWORK")
  border_top:SetAtlas("battlefieldminimap-border-top")
  border_top:SetPoint("TOPLEFT", border_topleft, "TOPRIGHT")
  border_top:SetPoint("BOTTOMRIGHT", border_topright, "BOTTOMLEFT")

  local border_bottom = frame:CreateTexture(nil, "ARTWORK")
  border_bottom:SetAtlas("battlefieldminimap-border-bottom")
  border_bottom:SetPoint("TOPLEFT", border_bottomleft, "TOPRIGHT")
  border_bottom:SetPoint("BOTTOMRIGHT", border_bottomright, "BOTTOMLEFT")

  local border_left = frame:CreateTexture(nil, "ARTWORK")
  border_left:SetAtlas("battlefieldminimap-border-left")
  border_left:SetPoint("TOPLEFT", border_topleft, "BOTTOMLEFT")
  border_left:SetPoint("BOTTOMRIGHT", border_bottomleft, "TOPRIGHT")

  local border_right = frame:CreateTexture(nil, "ARTWORK")
  border_right:SetAtlas("battlefieldminimap-border-right")
  border_right:SetPoint("TOPLEFT", border_topright, "BOTTOMLEFT")
  border_right:SetPoint("BOTTOMRIGHT", border_bottomright, "TOPRIGHT")

  frame.background = background
  frame.close_button = close_button

  return frame
end

function Widget.create_tab_button(optional_attrs, parent)
  local default_attrs = {
    text = ""
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Button", nil, parent)
  frame:SetFrameStrata("LOW")
  frame:SetSize(64, 32)

  local left = frame:CreateTexture(nil, "BACKGROUND")
  left:SetTexture(130942)
  left:SetTexCoord(0.0, 0.25, 0.0, 1.0)
  left:SetSize(16, 32)
  left:SetPoint("TOPLEFT")

  local middle = frame:CreateTexture(nil, "BACKGROUND")
  middle:SetTexture(130942)
  middle:SetTexCoord(0.25, 0.75, 0.0, 1.0)
  middle:SetSize(44, 32)
  middle:SetPoint("LEFT", left, "RIGHT")

  local right = frame:CreateTexture(nil, "BACKGROUND")
  right:SetTexture(130942)
  right:SetTexCoord(0.75, 1.0, 0.0, 1.0)
  right:SetSize(16, 32)
  right:SetPoint("LEFT", middle, "RIGHT")

  local highlight = frame:CreateTexture(nil, "BACKGROUND")
  highlight:SetTexture(136580)
  highlight:SetPoint("LEFT", left, 0, -7)
  highlight:SetPoint("RIGHT", right, 0, -7)
  highlight:SetBlendMode("ADD")

  local pushed = frame:CreateTexture(nil, "BACKGROUND")
  pushed:SetTexture(375515)
  pushed:SetPoint("LEFT", left, 0, -7)
  pushed:SetPoint("RIGHT", right, 0, -7)
  pushed:SetBlendMode("ADD")

  local font_string = frame:CreateFontString()
  font_string:SetPoint("LEFT", left, 8, -6)
  font_string:SetPoint("RIGHT", right, -8, -6)
  font_string:SetJustifyH("CENTER")

  frame:SetHighlightTexture(highlight)
  frame:SetPushedTexture(pushed)
  frame:SetPushedTextOffset(0, 0)
  frame:SetNormalFontObject("GameFontNormalSmall")
  frame:SetFontString(font_string)
  frame:SetText(attrs.text)

  return frame
end

function Widget.create_tab_panel(optional_attrs, parent)
  local default_attrs = {}
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", nil, parent)
  frame:Hide()

  return frame
end

function Widget.create_background(optional_attrs, parent)
  local default_attrs = {
    texture = 130871,
    color = {0, 0, 0, 1}
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", nil, parent)
  frame:SetAllPoints(parent)

  local background = frame:CreateTexture()
  background:SetAllPoints(frame)
  background:SetTexture(attrs.texture)
  background:SetVertexColor(unpack(attrs.color))

  return frame
end

function Widget.create_label(optional_attrs, parent)
  local default_attrs = {
    size = {80, 16},
    icon = nil,
    name = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", attrs.name, parent)
  frame:EnableMouse(true)
  frame:EnableMouseWheel(true)

  frame:SetMovable(true)
  frame:SetClampedToScreen(true)
  frame:SetFrameStrata("LOW")
  frame:SetFrameLevel(1)
  frame:SetSize(unpack(attrs.size))

  local icon = frame:CreateTexture()
  icon:SetTexture(attrs.icon)
  icon:SetSize(16, 16)
  icon:SetPoint("LEFT")

  local text = frame:CreateFontString()
  text:SetFont(STANDARD_TEXT_FONT, 12, "OUTLINE")
  text:SetAllPoints(frame)
  text:SetJustifyH("LEFT")
  text:SetPoint("TOPLEFT", 16, 0)
  text:SetPoint("BOTTOMRIGHT", 0, 0)

  local background = Widget.create_background({color = {0, 0, 0, 0.8}}, frame)
  background:Hide()
  background:SetFrameLevel(0)

  frame.icon = icon
  frame.text = text
  frame.background = background

  return frame
end

function Widget.create_button(optional_attrs, parent)
  local default_attrs = {
    text = "",
    size = {80, 24},
    tooltip = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate")
  frame:SetText(attrs.text)
  frame:SetSize(unpack(attrs.size))
  frame.tooltipText = attrs.tooltip

  return frame
end

function Widget.create_checkbox(optional_attrs, parent)
  local default_attrs = {
    text = "",
    checked = false
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("CheckButton", nil, parent, "UICheckButtonTemplate")
  frame:SetChecked(attrs.checked)
  frame.text:SetText(attrs.text)

  return frame
end

function Widget.create_editbox(optional_attrs, parent)
  local default_attrs = {
    text = "",
    size = {80, 24},
    auto_focus = false
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("EditBox", nil, parent, "InputBoxTemplate")

  frame:SetAutoFocus(attrs.auto_focus)
  frame:SetSize(unpack(attrs.size))
  frame:SetText(attrs.text)

  return frame
end

Package.Kernel = Kernel
Package.String = String
Package.Enum = Enum
Package.Utils = Utils
Package.Observable = Observable
Package.Widget = Widget
Package.L = Package[GetLocale()] or Package["enUS"]