local ADDON, Package = ...

local Module = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local omit = Package.Enum.omit
local is_member = Package.Enum.is_member
local camelcase = Package.String.camelcase
local purge_keyboard_focus = Package.Utils.purge_keyboard_focus
local create_character_icon = Package.Utils.create_character_icon
local create_character_name = Package.Utils.create_character_name
local create_window = Package.Widget.create_window
local create_tab_button = Package.Widget.create_tab_button
local create_tab_panel = Package.Widget.create_tab_panel
local create_button = Package.Widget.create_button
local create_checkbox = Package.Widget.create_checkbox
local create_editbox = Package.Widget.create_editbox

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PREFERENCES = {
  {name = "money", icon = 133784, module = Package.Money},
  {name = "durability", icon = 136241, module = Package.Durability},
  {name = "raid_information", icon = 134478, module = Package.RaidInformation},
  {name = "move_speed", icon = 134010, module = Package.MoveSpeed},
  {name = "location", icon = 134269, module = Package.Location},
  {name = "profession_cooldown", icon = 136106, module = Package.ProfessionCooldown}
}

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------
local function delete_realm(storage)
  table.wipe(storage)
end

local function delete_character(storage, name)
  local realm
  for key, value in pairs(storage) do
    if not is_member({"use", "position"}, key) then
      storage[key] = omit(value, {name}, 0)
    end
  end
end

local function format_decimal(value)
  return string.format("%.2f", value)
end

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local storage = nil

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local window = nil
local tabs = {}

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_window_drag_start(self) self:GetParent():StartMoving() end

function Module.on_window_drag_stop(self) self:GetParent():StopMovingOrSizing() end

function Module.on_window_mouse_up(self) purge_keyboard_focus() end

function Module.on_window_show()
  for at = 1, #PREFERENCES do
    PREFERENCES[at].module.hot()
  end
end

function Module.on_window_hide()
  for at = 1, #PREFERENCES do
    PREFERENCES[at].module.cold()
  end
end

function Module.on_tab_select(self)
  Module.select_tab(tabs, self:GetID())
end

function Module.on_use_checkbox_show(self)
  local preference = PREFERENCES[self:GetID()]

  self:SetChecked(storage.use[preference.name])
end

function Module.on_use_checkbox_click(self)
  local preference = PREFERENCES[self:GetID()]

  storage.use[preference.name] = self:GetChecked()

  if storage.use[preference.name] then
    preference.module.enable()
    preference.module.hot()
  else
    preference.module.disable()
    preference.module.cold()
  end
end

function Module.on_position_show(self)
  local preference = PREFERENCES[self:GetID()]

  self.x_editbox:SetEnabled(storage.use[preference.name])
  self.y_editbox:SetEnabled(storage.use[preference.name])

  self.x_editbox:SetText(format_decimal(storage.position[preference.name].x))
  self.y_editbox:SetText(format_decimal(storage.position[preference.name].y))
end

function Module.on_position_editbox_enter_pressed(self)
  local parent = self:GetParent()
  local preference = PREFERENCES[parent:GetID()]

  self:ClearFocus()

  Package.Observable.broadcast("position", preference.name, parent.x_editbox:GetNumber(), parent.y_editbox:GetNumber())
end

function Module.on_position_change(name, x, y)
  local child = tabs[2].panel.children[name]

  if child then
    child.x_editbox:SetText(format_decimal(x))
    child.y_editbox:SetText(format_decimal(y))
  end
end

function Module.on_delete_realm_button_click(self)
  StaticPopup_Show("MEDUSA_STAPLE_DELETE_RECORD", nil, nil, {title = GetRealmName(), action = delete_realm})
end

function Module.on_delete_character_button_click(self)
  StaticPopup_Show("MEDUSA_STAPLE_DELETE_RECORD", nil, nil, {title = self.character_card:GetText(), action = delete_character, arg1 = self.character_name})
end

function Module.select_tab(tabs, index)
  for at = 1, #tabs do
    tabs[at].button:SetButtonState("NORMAL", false)
    tabs[at].panel:Hide()

    tabs[index].button:SetButtonState("PUSHED", true)
    tabs[index].panel:Show()
  end
end

function Module.create_configuration_window()
  window = create_window({size = {240, 240}, name = ADDON .. "Config"}, UIParent)
  window:SetPoint("CENTER")
  window:SetScript("OnMouseUp", Module.on_window_mouse_up)
  window:SetScript("OnShow", Module.on_window_show)
  window:SetScript("OnHide", Module.on_window_hide)
  window:SetClampedToScreen(false)
end

function Module.create_configuration_tabs()
  local tabs_attrs = {
    {name = "general", text = L["general"]},
    {name = "position", text = L["position"]},
    {name = "record", text = L["record"]}
  }

  for at = 1, #tabs_attrs do
    local tab = {}

    tab.name = tabs_attrs[at].name

    tab.button = create_tab_button({text = tabs_attrs[at].text}, window)
    tab.button:SetID(at)
    tab.button:SetPoint("TOPLEFT", (at - 1) * 74, 36)
    tab.button:SetScript("OnDragStart", Module.on_window_drag_start)
    tab.button:SetScript("OnDragStop", Module.on_window_drag_stop)
    tab.button:RegisterForDrag("LeftButton")

    tab.panel = create_tab_panel(nil, window)
    tab.panel:SetID(at)
    tab.panel:SetAllPoints(window)

    tabs[at] = tab
  end

  for at = 1, #tabs do
    tabs[at].button:SetScript("OnClick", Module.on_tab_select)
  end
end

function Module.create_configuration_general_tab()
  local relative = nil
  local parent = tabs[1].panel

  for at = 1, #PREFERENCES do
    local preference = PREFERENCES[at]

    local use_checkbox = create_checkbox({text = L[preference.name]}, parent)
    use_checkbox:SetID(at)
    use_checkbox:SetScript("OnClick", Module.on_use_checkbox_click)
    use_checkbox:SetScript("OnShow", Module.on_use_checkbox_show)
    use_checkbox.text:SetWidth(parent:GetWidth() - 16)
    use_checkbox.text:SetJustifyH("LEFT")

    if not relative then
      use_checkbox:SetPoint("TOPLEFT", 8, -8)
    else
      use_checkbox:SetPoint("TOP", relative, "BOTTOM")
    end

    relative = use_checkbox
  end
end

function Module.create_configuration_position_tab()
  local relative = nil
  local parent = tabs[2].panel

  if not parent.children then
    parent.children = {}
  end

  for at = 1, #PREFERENCES do
    local preference = PREFERENCES[at]

    local frame = CreateFrame("Frame", nil, parent)
    frame:SetID(at)
    frame:SetSize(200, 24)
    frame:SetScript("OnShow", Module.on_position_show)

    local icon = frame:CreateTexture()
    icon:SetPoint("LEFT")
    icon:SetSize(24, 24)
    icon:SetTexture(preference.icon)

    local x_editbox = create_editbox(nil, frame)
    x_editbox:SetPoint("LEFT", icon, "RIGHT", 8, 0)
    x_editbox:SetScript("OnEnterPressed", Module.on_position_editbox_enter_pressed)

    local y_editbox = create_editbox(nil, frame)
    y_editbox:SetPoint("LEFT", x_editbox, "RIGHT", 8, 0)
    y_editbox:SetScript("OnEnterPressed", Module.on_position_editbox_enter_pressed)

    if not relative then
      frame:SetPoint("TOPLEFT", 8, -8)
    else
      frame:SetPoint("TOP", relative, "BOTTOM", 0, -8)
    end

    frame.icon = icon
    frame.x_editbox = x_editbox
    frame.y_editbox = y_editbox

    relative = frame

    parent.children[preference.name] = frame
  end

  Package.Observable.subscribe("position", Module.on_position_change)
end

function Module.create_configuration_record_tab()
  local relative = nil
  local parent = tabs[3].panel

  for owner, state in pairs(storage.character) do
    local frame = CreateFrame("Button", nil, parent)
    frame:SetSize(128, 18)
    frame:SetNormalTexture(131071, "BLEND")
    frame:GetNormalTexture():SetVertexColor(0, 0, 0, 0.8)
    frame:SetHighlightTexture(130871, "ADD")
    frame:GetHighlightTexture():SetVertexColor(0, 1, 1, 0.1)
    frame:SetScript("OnClick", Module.on_delete_character_button_click)

    local character_card = frame:CreateFontString()
    character_card:SetFont(STANDARD_TEXT_FONT, 12, "OUTLINE")
    character_card:SetText(create_character_icon(state) .. " " .. create_character_name(state))
    character_card:SetPoint("LEFT", frame)

    local delete_icon = frame:CreateTexture(nil, "ARTWORK")
    delete_icon:SetSize(16, 16)
    delete_icon:SetTexture(851905)
    delete_icon:SetVertexColor(1, 0, 0)
    delete_icon:SetPoint("LEFT", frame, "RIGHT", -16, 0)

    if not relative then
      frame:SetPoint("TOPLEFT", 8, -8)
    else
      frame:SetPoint("TOP", relative, "BOTTOM", 0, 0)
    end

    frame.character_name = state.name
    frame.character_card = character_card
    frame.delete_icon = delete_icon

    relative = frame
  end

  local delete_realm_button = create_button({text = L["reset"]}, parent)
  delete_realm_button:SetPoint("BOTTOM", 0, 8)
  delete_realm_button:SetScript("OnClick", Module.on_delete_realm_button_click)
end

function Module.run()
  local realm = GetRealmName()

  storage = MedusaStapleStorage.realm[realm]

  Module.create_configuration_window()
  Module.create_configuration_tabs()
  Module.create_configuration_general_tab()
  Module.create_configuration_position_tab()
  Module.create_configuration_record_tab()

  Module.select_tab(tabs, 1)
end

--------------------------------------------------------------------------------
-- Static Popup Dialogs
--------------------------------------------------------------------------------
StaticPopupDialogs.MEDUSA_STAPLE_DELETE_RECORD = {
  button1 = YES,
  button2 = NO,
  hideOnEscape = true,
  whileDead = true,
  timeout = 0,
  text = L["ask_to_delete_record"],
  subText = "",
  OnShow = function (self, data)
    self.SubText:SetText(data.title)
  end,
  OnAccept = function (self, data)
    if data.action then
      data.action(storage, data.arg1)

      C_UI.Reload()
    end
  end
}

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.Config = Module
