local ADDON, Package = ...

local Module = {}
local Registry = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local create_label = Package.Widget.create_label

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PROXY = CreateFrame("Frame")

local COLORS = {
  WHITE = "ffffffff",
  YELLOW = "fff2d22e",
  ORANGE = "fff27405",
  RED = "fff2220f"
}

local ITEM_EQUIP_LOCS = {
  INVTYPE_HEAD = INVTYPE_HEAD,
  INVTYPE_SHOULDER = INVTYPE_SHOULDER,
  INVTYPE_CHEST = INVTYPE_CHEST,
  INVTYPE_ROBE = INVTYPE_ROBE,
  INVTYPE_WAIST = INVTYPE_WAIST,
  INVTYPE_LEGS = INVTYPE_LEGS,
  INVTYPE_FEET  = INVTYPE_FEET,
  INVTYPE_WRIST = INVTYPE_WRIST,
  INVTYPE_HAND = INVTYPE_HAND,
  INVTYPE_WEAPON = INVTYPE_WEAPON,
  INVTYPE_SHIELD = INVTYPE_SHIELD,
  INVTYPE_2HWEAPON = INVTYPE_2HWEAPON,
  INVTYPE_WEAPONMAINHAND = INVTYPE_WEAPONMAINHAND,
  INVTYPE_WEAPONOFFHAND = INVTYPE_WEAPONOFFHAND,
  INVTYPE_HOLDABLE = INVTYPE_HOLDABLE,
  INVTYPE_RANGED = INVTYPE_RANGED,
  INVTYPE_RANGEDRIGHT = INVTYPE_RANGEDRIGHT or INVTYPE_RANGED
}

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------
local function get_inventory_item_durability()
  local low = 1
  local high = 1
  local slots = {}

  for at = 0, 19 do
    local item_id = GetInventoryItemID("player", at)
    local item_name, _, item_quality, _, _, _, _, _, item_equip_loc, item_texture = GetItemInfo(item_id or 0)
    local item_low, item_high = GetInventoryItemDurability(at)

    if item_id and item_name and item_quality and item_equip_loc and item_texture and item_low and item_high then
      if item_low ~= item_high then

        low = low + item_low
        high = high + item_high

        table.insert(slots, {
          icon = CreateTextureMarkup(item_texture, 64, 64, 16, 16, 0, 1, 0, 1),
          slot = ITEM_EQUIP_LOCS[item_equip_loc],
          name = ITEM_QUALITY_COLORS[item_quality].color:WrapTextInColorCode(item_name),
          percent = math.floor(item_low / item_high * 100)
        })
      end
    end
  end

  return {name = "durability_inventory", low = low, high = high, percent = math.floor(low / high * 100), slots = slots}
end

local function get_container_item_durability()
  local low = 1
  local high = 1
  local slots = {}

  for bag = 0, NUM_BAG_FRAMES do
    for slot = 1, GetContainerNumSlots(bag) do
      local item_id = GetContainerItemID(bag, slot)
      local item_name, _, item_quality, _, _, _, _, _, item_equip_loc, item_texture = GetItemInfo(item_id or 0)
      local item_low, item_high = GetContainerItemDurability(bag, slot)

      if item_id and item_name and item_quality and item_equip_loc and item_texture and item_low and item_high then
        if item_low ~= item_high then
          low = low + item_low
          high = high + item_high

          table.insert(slots, {
            icon = CreateTextureMarkup(item_texture, 64, 64, 16, 16, 0, 1, 0, 1),
            slot = ITEM_EQUIP_LOCS[item_equip_loc],
            name = ITEM_QUALITY_COLORS[item_quality].color:WrapTextInColorCode(item_name),
            percent = math.floor(item_low / item_high * 100)
          })
        end
      end
    end
  end

  return {name = "durability_container", low = low, high = high, percent = math.floor(low / high * 100), slots = slots}
end

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------
local function get_alert_color(percent)
  if percent <= 25 then
    return COLORS.RED
  elseif percent <= 50 then
    return COLORS.ORANGE
  elseif percent <= 75 then
    return COLORS.YELLOW
  else
    return COLORS.WHITE
  end
end

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local position = {x = 0, y = 0}
local sources = {
  inventory = {},
  container = {}
}
local percent = 0

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local label = create_label({size = {64, 16}, icon = 136241}, UIParent)

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_label_position_change(name, x, y)
  if name == "durability" then
    position.x = x
    position.y = y

    label:ClearAllPoints()
    label:SetPoint("BOTTOMLEFT", x, y)
  end
end

function Module.on_label_drag_start(self)
  self:StartMoving()
end

function Module.on_label_drag_stop(self)
  self:StopMovingOrSizing()

  Package.Observable.broadcast("position", "durability", self:GetLeft(), self:GetBottom())
end

function Module.on_label_enter(self)
  GameTooltip:SetOwner(self, "ANCHOR_BOTTOM")
  GameTooltip:ClearLines()

  GameTooltip:AddLine(L["durability"])

  for _, source in pairs(sources) do
    GameTooltip:AddLine(" ")
    GameTooltip:AddDoubleLine(L[source.name], WrapTextInColorCode(source.percent .. "%", get_alert_color(source.percent)), 1, 1, 1)

    for _, item in pairs(source.slots) do
      GameTooltip:AddDoubleLine(item.icon .. " " .. item.name, WrapTextInColorCode(item.percent .. "%", get_alert_color(item.percent)), 1, 1, 1)
    end
  end

  GameTooltip:Show()
end

function Module.on_label_leave(self)
  GameTooltip:Hide()
end

function Module.update_status()
  sources.inventory = get_inventory_item_durability()
  sources.container = get_container_item_durability()
  percent = math.floor((sources.inventory.low + sources.container.low) / (sources.inventory.high + sources.container.high) * 100)

  label.text:SetFormattedText(WrapTextInColorCode(percent .. "%%", get_alert_color(percent)))
end

function Module.enable()
  local realm = GetRealmName()
  local storage = MedusaStapleStorage.realm[realm]

  position = storage.position.durability

  label:RegisterForDrag("LeftButton")
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label:SetPoint("BOTTOMLEFT", position.x, position.y)
  label:Show()

  PROXY:RegisterEvent("UPDATE_INVENTORY_DURABILITY")
  PROXY:SetScript("OnEvent", Module.on_event)

  Package.Observable.subscribe("position", Module.on_label_position_change)

  Module:update_status()
end

function Module.disable()
  label:Hide()

  PROXY:UnregisterEvent("UPDATE_INVENTORY_DURABILITY")
  PROXY:SetScript("OnEvent", nil)

  Package.Observable.unsubscribe("position", Module.on_label_position_change)
end

function Module.hot()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", Module.on_label_drag_start)
  label:SetScript("OnDragStop", Module.on_label_drag_stop)
  label:SetScript("OnEnter", nil)
  label:SetScript("OnLeave", nil)

  label.background:Show()
end

function Module.cold()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", nil)
  label:SetScript("OnDragStop", nil)
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label.background:Hide()
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

--------------------------------------------------------------------------------
-- Registry
--------------------------------------------------------------------------------
function Registry.UPDATE_INVENTORY_DURABILITY()
  Module.update_status()
end

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.Durability = Module