local ADDON, Package = ...

local Module = {}
local Registry = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local create_label = Package.Widget.create_label
local create_character_icon = Package.Utils.create_character_icon
local create_character_name = Package.Utils.create_character_name

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PROXY = CreateFrame("Frame")

local HEARTHSTONE_SPELL_ID = 8690

local COLORS = {
  WHITE = CreateColor(1, 1, 1),
  GRAY = CreateColor(0.5, 0.5, 0.5),
  GREEN = CreateColor(0, 1, 0),
  CYAN = CreateColor(0, 1, 1)
}

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------
local function get_spell_cooldown(spell_id)
  local elapsed = GetTime()
  local started, duration = GetSpellCooldown(spell_id)
  local left = started + duration - elapsed

  if started > 0 then
    if left > duration then
      left = left - (2^32 / 1000)
    end

    return left
  end

  return 0
end

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------
local function format_zone_and_subzone(zone, subzone)
  if zone ~= subzone and zone ~= nil and subzone ~= "" and subzone ~= nil then
    return string.format("%s - %s", zone, subzone)
  elseif zone == "" or zone == nil then
    return UNKNOWN
  else
    return zone
  end
end

local function format_position(x, y)
  return COLORS.GRAY:WrapTextInColorCode(string.format("%.2f", x * 100)) .. "•" .. COLORS.GRAY:WrapTextInColorCode(string.format("%.2f", y * 100))
end

local function format_hearthstone_bind(bind)
  if bind == "" or bind == nil then
    return COLORS.WHITE:WrapTextInColorCode(UNKNOWN .. ": ")
  else
    return COLORS.WHITE:WrapTextInColorCode(bind .. ": ")
  end
end

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local time_since_last = 0
local cache = {}
local position = {x = 0, y = 0}
local character = {}
local location = {}

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local label = create_label({size = {104, 16}, icon = 134269}, UIParent)

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_label_position_change(name, x, y)
  if name == "location" then
    position.x = x
    position.y = y

    label:ClearAllPoints()
    label:SetPoint("BOTTOMLEFT", x, y)
  end
end

function Module.on_label_drag_start(self)
  self:StartMoving()
end

function Module.on_label_drag_stop(self)
  self:StopMovingOrSizing()

  Package.Observable.broadcast("position", "location", self:GetLeft(), self:GetBottom())
end

function Module.on_label_enter(self)
  local player = UnitName("player")
  local timestamp = GetServerTime()

  for owner, state in pairs(location) do
    if not cache[owner] then
      local character_card = create_character_icon(character[owner]) .. " " .. create_character_name(character[owner])
      local location_zone = format_zone_and_subzone(state.zone, state.subzone)
      local location_position = format_position(state.position.x, state.position.y)
      local location_hearthstone_bind = format_hearthstone_bind(state.hearthstone.bind)
      local location_hearthstone_timestamp = state.hearthstone.timestamp

      cache[owner] = {
        character_card = character_card,
        location_zone = location_zone,
        location_position = location_position,
        location_hearthstone_bind = location_hearthstone_bind,
        location_hearthstone_timestamp = location_hearthstone_timestamp
      }
    end
  end

  GameTooltip:SetOwner(self, "ANCHOR_BOTTOM")
  GameTooltip:ClearLines()

  GameTooltip:AddLine(L["location"] .. " " .. COLORS.CYAN:WrapTextInColorCode("<Alt>"))
  GameTooltip:AddLine(" ")

  for _, state in pairs(cache) do
    if IsAltKeyDown() then
      GameTooltip:AddDoubleLine(state.character_card, state.location_zone .. ": " .. state.location_position, nil, nil, nil, 1, 1, 1)
    else
      if state.location_hearthstone_timestamp - timestamp <= 0 then
        GameTooltip:AddDoubleLine(state.character_card, state.location_hearthstone_bind .. COLORS.GREEN:WrapTextInColorCode(L["usable"]))
      else
        GameTooltip:AddDoubleLine(state.character_card, state.location_hearthstone_bind .. SecondsToTime(state.location_hearthstone_timestamp - timestamp), nil, nil, nil, 1, 1, 1)
      end
    end
  end

  cache[player] = nil

  GameTooltip:Show()
end

function Module.on_label_leave(self)
  GameTooltip:Hide()
end

function Module.update_zone()
  local player = UnitName("player")

  location[player].zone = GetZoneText()
  location[player].subzone = GetSubZoneText()
end

function Module.update_position()
  local player = UnitName("player")

  local map = C_Map.GetBestMapForUnit("player")
  local map_position = nil

  if map then
    map_position = C_Map.GetPlayerMapPosition(map, "player")

    if map_position then
      location[player].position.x = map_position.x
      location[player].position.y = map_position.y

      label.text:SetFormattedText("%.2f•%.2f", map_position.x * 100, map_position.y * 100)
    end
  end

  if not map or not map_position then
    location[player].position.x = 0
    location[player].position.y = 0

    label.text:SetFormattedText(L["location_unavailable"])
  end
end

function Module.update_hearthstone()
  local player = UnitName("player")
  local timestamp = GetServerTime()

  location[player].hearthstone.bind = GetBindLocation()
  location[player].hearthstone.timestamp = timestamp + get_spell_cooldown(HEARTHSTONE_SPELL_ID)
end

function Module.enable()
  local realm = GetRealmName()
  local storage = MedusaStapleStorage.realm[realm]

  position = storage.position.location
  character = storage.character
  location = storage.location

  label:RegisterForDrag("LeftButton")
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label:SetPoint("BOTTOMLEFT", position.x, position.y)
  label:Show()

  PROXY:RegisterEvent("MODIFIER_STATE_CHANGED")
  PROXY:RegisterEvent("ZONE_CHANGED")
  PROXY:RegisterEvent("ZONE_CHANGED_INDOORS")
  PROXY:RegisterEvent("ZONE_CHANGED_NEW_AREA")
  PROXY:RegisterEvent("HEARTHSTONE_BOUND")
  PROXY:RegisterUnitEvent("UNIT_SPELLCAST_SUCCEEDED", "player")
  PROXY:SetScript("OnEvent", Module.on_event)
  PROXY:SetScript("OnUpdate", Module.on_update)

  Package.Observable.subscribe("position", Module.on_label_position_change)

  Module.update_zone()
  Module.update_position()
  Module.update_hearthstone()
end

function Module.disable()
  label:Hide()

  PROXY:UnregisterEvent("MODIFIER_STATE_CHANGED")
  PROXY:UnregisterEvent("ZONE_CHANGED")
  PROXY:UnregisterEvent("ZONE_CHANGED_INDOORS")
  PROXY:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
  PROXY:UnregisterEvent("HEARTHSTONE_BOUND")
  PROXY:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
  PROXY:SetScript("OnEvent", nil)
  PROXY:SetScript("OnUpdate", nil)

  Package.Observable.unsubscribe("position", Module.on_label_position_change)
end

function Module.hot()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", Module.on_label_drag_start)
  label:SetScript("OnDragStop", Module.on_label_drag_stop)
  label:SetScript("OnEnter", nil)
  label:SetScript("OnLeave", nil)

  label.background:Show()
end

function Module.cold()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", nil)
  label:SetScript("OnDragStop", nil)
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label.background:Hide()
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

function Module.on_update(self, elapsed)
  time_since_last = time_since_last + elapsed

  if time_since_last > 4.0 then
    time_since_last = 0

    Module.update_position()
  end
end

--------------------------------------------------------------------------------
-- Registry
--------------------------------------------------------------------------------
function Registry.MODIFIER_STATE_CHANGED(self, key, state)
  if label:IsShown() and label:IsMouseOver() and key == "LALT" then
    Module.on_label_enter(label)
  end
end

function Registry.ZONE_CHANGED()
  Module.update_zone()
end

function Registry.ZONE_CHANGED_INDOORS()
  Module.update_zone()
end

function Registry.ZONE_CHANGED_NEW_AREA()
  Module.update_zone()
end

function Registry.HEARTHSTONE_BOUND()
  Module.update_hearthstone()
end

function Registry.UNIT_SPELLCAST_SUCCEEDED(self, unit_target, cast_guid, spell_id)
  if unit_target == "player" and spell_id == HEARTHSTONE_SPELL_ID then
    C_Timer.After(1, Module.update_hearthstone)
  end
end

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.Location = Module