local ADDON, Package = ...

local Module = {}
local Registry = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local create_label = Package.Widget.create_label
local create_character_icon = Package.Utils.create_character_icon
local create_character_name = Package.Utils.create_character_name

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PROXY = CreateFrame("Frame")

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------
local function get_coin_texture_string(amount)
  local stash = {}
  local g = floor(amount / (COPPER_PER_SILVER * SILVER_PER_GOLD))
  local s = floor((amount - (g * COPPER_PER_SILVER * SILVER_PER_GOLD)) / COPPER_PER_SILVER)
  local c = mod(amount, COPPER_PER_SILVER)

  if g > 0 then
    table.insert(stash, string.format("|cfff6d144%s%s|r", FormatLargeNumber(g), GOLD_AMOUNT_SYMBOL))
  end

  if s > 0 then
    table.insert(stash, string.format("|cffb0aeb1%s%s|r", s, SILVER_AMOUNT_SYMBOL))
  end

  if c > 0 or (g == 0 and s == 0) then
    table.insert(stash, string.format("|cffbf744d%s%s|r", c, COPPER_AMOUNT_SYMBOL))
  end

  return table.concat(stash, " ")
end

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local cache = {}
local position = {x = 0, y = 0}
local character = {}
local money = {}

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local label = create_label({size = {148, 16}, icon = 133784}, UIParent)

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_label_position_change(name, x, y)
  if name == "money" then
    position.x = x
    position.y = y

    label:ClearAllPoints()
    label:SetPoint("BOTTOMLEFT", x, y)
  end
end

function Module.on_label_drag_start(self)
  self:StartMoving()
end

function Module.on_label_drag_stop(self)
  self:StopMovingOrSizing()

  Package.Observable.broadcast("position", "money", self:GetLeft(), self:GetBottom())
end

function Module.on_label_enter(self)
  local player = UnitName("player")
  local assets = 0

  for owner, state in pairs(money) do
    if not cache[owner] then
      local character_card = create_character_icon(character[owner]) .. " " .. create_character_name(character[owner])
      local money_amount = GetMoneyString(state.amount, true)

      cache[owner] = {
        character_card = character_card,
        money_amount = money_amount
      }
    end

    assets = assets + state.amount
  end

  GameTooltip:SetOwner(self, "ANCHOR_BOTTOM")
  GameTooltip:ClearLines()

  GameTooltip:AddDoubleLine(L["money"], GetMoneyString(assets, true), nil, nil, nil, 1, 1, 1)
  GameTooltip:AddLine(" ")

  for _, state in pairs(cache) do
    GameTooltip:AddDoubleLine(state.character_card, state.money_amount, nil, nil, nil, 1, 1, 1)
  end

  cache[player] = nil

  GameTooltip:Show()
end

function Module.on_label_leave(self)
  GameTooltip:Hide()
end

function Module:update_amount()
  local player = UnitName("player")
  local amount = GetMoney()

  money[player].amount = amount

  label.text:SetFormattedText(get_coin_texture_string(amount))
end

function Module.enable()
  local realm = GetRealmName()
  local storage = MedusaStapleStorage.realm[realm]

  position = storage.position.money
  character = storage.character
  money = storage.money

  label:RegisterForDrag("LeftButton")
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label:SetPoint("BOTTOMLEFT", position.x, position.y)
  label:Show()

  PROXY:RegisterEvent("PLAYER_MONEY")
  PROXY:SetScript("OnEvent", Module.on_event)

  Package.Observable.subscribe("position", Module.on_label_position_change)

  Module.update_amount()
end

function Module.disable()
  label:Hide()

  PROXY:UnregisterEvent("PLAYER_MONEY")
  PROXY:SetScript("OnEvent", nil)

  Package.Observable.unsubscribe("position", Module.on_label_position_change)
end

function Module.hot()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", Module.on_label_drag_start)
  label:SetScript("OnDragStop", Module.on_label_drag_stop)
  label:SetScript("OnEnter", nil)
  label:SetScript("OnLeave", nil)

  label.background:Show()
end

function Module.cold()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", nil)
  label:SetScript("OnDragStop", nil)
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label.background:Hide()
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

--------------------------------------------------------------------------------
-- Registry
--------------------------------------------------------------------------------
function Registry.PLAYER_MONEY()
  Module.update_amount()
end

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.Money = Module
