local ADDON, Package = ...

local Module = {}
local Registry = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local create_label = Package.Widget.create_label
local create_character_icon = Package.Utils.create_character_icon
local create_character_name = Package.Utils.create_character_name

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PROXY = CreateFrame("Frame")

local PROFESSION_TYPES = {
  LEATHERWORKING = 32549, -- Master
  TAILORING = 26790, -- Master
  ALCHEMY = 28596, -- Master
  JEWELCRAFTING = 28897 -- Master
}

local PROFESSION_SPELLS = {
  LEATHERWORKING = { -- Leatherworking 133611
    [19566] = 132836 -- Salt Shaker
  },
  TAILORING = { -- Tailoring 136249
    [26751] = 132897, -- Primal Mooncloth
    [31373] = 132910, -- Spellcloth 
    [36686] = 132887, -- Shadowcloth
  },
  ALCHEMY = { -- Alchemy 136240
    [17559] = 135830, -- Transmute: Air to Fire
    [17560] = 136102, -- Transmute: Fire to Earth
    [17561] = 136007, -- Transmute: Earth to Water
    [17562] = 136022, -- Transmute: Water to Air
    [17563] = 136007, -- Transmute: Undeath to Water
    [17564] = 136195, -- Transmute: Water to Undeath
    [17565] = 136102, -- Transmute: Life to Earth
    [17566] = 136006, -- Transmute: Earth to Life
    [11479] = 133217, -- Transmute: Iron to Gold
    [11480] = 133222, -- Transmute: Mithril to Truesilver

    [28566] = 132847, -- Transmute: Primal Air to Fire
    [28567] = 132852, -- Transmute: Primal Earth to Water
    [28568] = 132846, -- Transmute: Primal Fire to Earth
    [28569] = 132845, -- Transmute: Primal Water to Air
    [28580] = 132852, -- Transmute: Primal Shadow to Water
    [28581] = 132851, -- Transmute: Primal Water to Shadow
    [28582] = 132847, -- Transmute: Primal Mana to Fire
    [28583] = 132849, -- Transmute: Primal Fire to Mana
    [28584] = 132846, -- Transmute: Primal Life to Earth
    [28585] = 132848, -- Transmute: Primal Earth to Life
    [29688] = 136050, -- Transmute: Primal Might
    [32765] = 134097, -- Transmute: Earthstorm Diamond
    [32766] = 134098, -- Transmute: Skyfire Diamond
  },
  JEWELCRAFTING = { -- Jewelcrafting 134072
    [47280] = 134096, -- Brilliant Glass
  }
}

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------
local function get_spell_cooldown(spell_id)
  local elapsed = GetTime()
  local started, duration = GetSpellCooldown(spell_id)
  local left = started + duration - elapsed

  if started > 0 then
    if left > duration then
      left = left - (2^32 / 1000)
    end

    return left
  end

  return 0
end

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------
function to_time_from_seconds(seconds)
  return format("%02d:%02d", (seconds / 3600), (seconds % 3600) / 60)
end

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local time_since_last = 0
local cache = {}
local position = {x = 0, y = 0}
local character = {}
local profession_cooldown = {}

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local label = create_label({size = {64, 16}, icon = 136106}, UIParent)

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_label_position_change(name, x, y)
  if name == "profession_cooldown" then
    position.x = x
    position.y = y

    label:ClearAllPoints()
    label:SetPoint("BOTTOMLEFT", x, y)
  end
end

function Module.on_label_drag_start(self)
  self:StartMoving()
end

function Module.on_label_drag_stop(self)
  self:StopMovingOrSizing()

  Package.Observable.broadcast("position", "profession_cooldown", self:GetLeft(), self:GetBottom())
end

function Module.on_label_enter(self)
  local player = UnitName("player")
  local timestamp = GetServerTime()

  for owner, state in pairs(profession_cooldown) do
    if not cache[owner] then
      local character_card = create_character_icon(character[owner]) .. " " .. create_character_name(character[owner])
      local profession_cooldown_spells = {}

      for spell_type, spells in pairs(state) do
        for spell_id, spell_timestamp in pairs(spells) do
          local spell_texture = PROFESSION_SPELLS[spell_type][spell_id]

          if spell_type == "ALCHEMY" then
            spell_texture = 136240
          end

          local spell_icon = CreateTextureMarkup(spell_texture, 64, 64, 16, 16, 0, 1, 0, 1)

          table.insert(profession_cooldown_spells, {
            icon = spell_icon,
            timestamp = spell_timestamp
          })
        end
      end

      cache[owner] = {
        character_card = character_card,
        profession_cooldown_spells = profession_cooldown_spells
      }
    end
  end

  GameTooltip:SetOwner(self, "ANCHOR_BOTTOM")
  GameTooltip:ClearLines()

  GameTooltip:AddLine(L["profession_cooldown"])
  GameTooltip:AddLine(" ")

  for _, state in pairs(cache) do
    if #state.profession_cooldown_spells > 0 then
      local spell_cards = ""

      for _, spell in pairs(state.profession_cooldown_spells) do
        local spell_card = spell.icon
        local spell_cooldown = spell.timestamp - timestamp

        if spell_cooldown <= 0 then
          spell_cards = spell_cards .. spell_card .. " " .. GREEN_FONT_COLOR:WrapTextInColorCode("00:00") .. " "
        else
          spell_cards = spell_cards .. spell_card .. " " .. to_time_from_seconds(spell_cooldown) .. " "
        end
      end

      GameTooltip:AddDoubleLine(state.character_card, spell_cards:trim(), 1, 1, 1, 1, 1, 1)
    end
  end

  cache[player] = nil

  GameTooltip:Show()
end

function Module.on_label_leave(self)
  GameTooltip:Hide()
end

function Module.update_spells()
  local player = UnitName("player")
  local timestamp = GetServerTime()

  for spell_type, spell_ids in pairs(PROFESSION_SPELLS) do
    local spell_count = 0

    if IsPlayerSpell(PROFESSION_TYPES[spell_type]) then
      for spell_id, _ in pairs(spell_ids) do
        if spell_type == "TAILORING" and spell_count < 4 then
          profession_cooldown[player][spell_type][spell_id] = timestamp + get_spell_cooldown(spell_id)
        end

        if spell_type ~= "TAILORING" and spell_count < 1 then
          profession_cooldown[player][spell_type][spell_id] = timestamp + get_spell_cooldown(spell_id)
        end

        spell_count = spell_count + 1
      end
    end

    spell_count = 0
  end
end

function Module.update_status()
  local player = UnitName("player")
  local timestamp = GetServerTime()
  local queued = 0
  local usable = 0

  for _, state in pairs(profession_cooldown) do
    for spell_type, spells in pairs(state) do
      for _, spell_timestamp in pairs(spells) do
        local spell_cooldown = spell_timestamp - timestamp

        if spell_cooldown <= 0 then
          usable = usable + 1
        end

        queued = queued + 1
      end
    end
  end

  label.text:SetFormattedText(usable > 0 and GREEN_FONT_COLOR:WrapTextInColorCode("%s/%s") or "%s/%s", usable, queued)
end

function Module.enable()
  local realm = GetRealmName()
  local storage = MedusaStapleStorage.realm[realm]

  position = storage.position.profession_cooldown
  character = storage.character
  profession_cooldown = storage.profession_cooldown

  label:RegisterForDrag("LeftButton")
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label:SetPoint("BOTTOMLEFT", position.x, position.y)
  label:Show()

  PROXY:RegisterUnitEvent("UNIT_SPELLCAST_SUCCEEDED", "player")
  PROXY:SetScript("OnEvent", Module.on_event)
  PROXY:SetScript("OnUpdate", Module.on_update)

  Package.Observable.subscribe("position", Module.on_label_position_change)

  Module.update_spells()
  Module.update_status()
end

function Module.disable()
  label:Hide()

  PROXY:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
  PROXY:SetScript("OnEvent", nil)
  PROXY:SetScript("OnUpdate", nil)

  Package.Observable.unsubscribe("position", Module.on_label_position_change)
end

function Module.hot()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", Module.on_label_drag_start)
  label:SetScript("OnDragStop", Module.on_label_drag_stop)
  label:SetScript("OnEnter", nil)
  label:SetScript("OnLeave", nil)

  label.background:Show()
end

function Module.cold()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", nil)
  label:SetScript("OnDragStop", nil)
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label.background:Hide()
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

function Module.on_update(self, elapsed)
  time_since_last = time_since_last + elapsed

  if time_since_last > 60.0 then
    time_since_last = 0

    Module.update_status()
  end
end

--------------------------------------------------------------------------------
-- Registry
--------------------------------------------------------------------------------
function Registry.UNIT_SPELLCAST_SUCCEEDED(self, unit_target, cast_guid, spell_id)
  if unit_target == "player" then
    if PROFESSION_SPELLS.LEATHERWORKING[spell_id] or PROFESSION_SPELLS.TAILORING[spell_id] or PROFESSION_SPELLS.ALCHEMY[spell_id] then
      C_Timer.After(1, Module.update_spells)
      C_Timer.After(1, Module.update_status)
    end
  end
end

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.ProfessionCooldown = Module