local ADDON, Package = ...

local Module = {}
local Registry = {}

--------------------------------------------------------------------------------
-- Medusa
--------------------------------------------------------------------------------
local L = Package.L
local create_label = Package.Widget.create_label
local create_character_icon = Package.Utils.create_character_icon
local create_character_name = Package.Utils.create_character_name

--------------------------------------------------------------------------------
-- Predefine
--------------------------------------------------------------------------------
local PROXY = CreateFrame("Frame")

local COLORS = {
  CYAN = CreateColor(0, 1, 1)
}

local RAID_ZONES = {249, 309, 409, 469, 509, 531, 532, 533, 544, 565}

local RAID_QUESTS = {[6502] = true, [6602] = true, [7848] = true, [7761] = true, [9837] = true}

local RAID_REQUIREMENTS = {
  [249] = {6502, 6602},
  [409] = {7848},
  [469] = {7761},
  [532] = {9837}
}

local MIN_RAID_LEVEL = 68

local MAX_RAID_INSTANCES = #RAID_ZONES

--------------------------------------------------------------------------------
-- Reform
--------------------------------------------------------------------------------
function get_saved_instance_info()
  local timestamp = GetServerTime()
  local stash = {}

  for at = 1, GetNumSavedInstances() do
    local name, id, left, _, _, _, _, _, is_raid, max_players, max_bosses, min_bosses = GetSavedInstanceInfo(at)

    if is_raid then
      stash[name] = {
        id = id,
        name = name,
        timestamp = timestamp + left,
        max_players = max_players,
        max_bosses = max_bosses,
        min_bosses = min_bosses
      }
    end
  end

  return stash
end

--------------------------------------------------------------------------------
-- Utility
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- State
--------------------------------------------------------------------------------
local cache = {}
local position = {x = 0, y = 0}
local character = {}
local raid_information = {}

--------------------------------------------------------------------------------
-- Object
--------------------------------------------------------------------------------
local label = create_label({size = {64, 16}, icon = 134478}, UIParent)

--------------------------------------------------------------------------------
-- Module
--------------------------------------------------------------------------------
function Module.on_label_position_change(name, x, y)
  if name == "raid_information" then
    position.x = x
    position.y = y

    label:ClearAllPoints()
    label:SetPoint("BOTTOMLEFT", x, y)
  end
end

function Module.on_label_drag_start(self)
  self:StartMoving()
end

function Module.on_label_drag_stop(self)
  self:StopMovingOrSizing()

  Package.Observable.broadcast("position", "raid_information", self:GetLeft(), self:GetBottom())
end

function Module.on_label_enter(self)
  local player = UnitName("player")
  local timestamp = GetServerTime()

  for owner, state in pairs(raid_information) do
    if not cache[owner] then
      local character_card = create_character_icon(character[owner]) .. " " .. create_character_name(character[owner])
      local raid_information_quests = {}
      local raid_information_instances = {}

      for instance_name, is_quest_completed in pairs(state.quests) do
        if not is_quest_completed and character[owner].level >= MIN_RAID_LEVEL then
          table.insert(raid_information_quests, {
            name = instance_name,
            requirement = REQUIRES_LABEL .. " " .. QUESTS_LABEL
          })
        end
      end

      for instance_name, instance in pairs(state.instances) do
        table.insert(raid_information_instances, {
          id = instance.id,
          name = instance.name,
          progress = instance.min_bosses .. "/" .. instance.max_bosses,
          timestamp = instance.timestamp,
          max_players = instance.max_players
        })
      end

      cache[owner] = {
        character_card = character_card,
        raid_information_quests = raid_information_quests,
        raid_information_instances = raid_information_instances
      }
    end
  end

  GameTooltip:SetOwner(self, "ANCHOR_BOTTOM")
  GameTooltip:ClearLines()

  GameTooltip:AddLine(L["raid_information"] .. " " .. COLORS.CYAN:WrapTextInColorCode("<Alt>"))

  for _, state in pairs(cache) do
    if #state.raid_information_quests > 0 or #state.raid_information_instances > 0  then
      GameTooltip:AddLine(" ")
      GameTooltip:AddLine(state.character_card)
    end

    for _, instance in pairs(state.raid_information_quests) do
      GameTooltip:AddDoubleLine(instance.name, instance.requirement, 0.251, 0.753, 0.251, 0.8, 0.1, 0.1)
    end

    for _, instance in pairs(state.raid_information_instances) do
      if instance.timestamp - timestamp > 0 then
        if IsAltKeyDown() then
          GameTooltip:AddDoubleLine(instance.name .. " " .. instance.progress, instance.id, 0.251, 0.753, 0.251, 1, 1, 1)
        else
          GameTooltip:AddDoubleLine(instance.name .. " " .. instance.max_players, SecondsToTime(instance.timestamp - timestamp), 0.251, 0.753, 0.251, 1, 1, 1)
        end
      end
    end
  end

  cache[player] = nil

  GameTooltip:Show()
end

function Module.on_label_leave(self)
  GameTooltip:Hide()
end

function Module.update_instances()
  local player = UnitName("player")

  raid_information[player].instances = get_saved_instance_info()
end

function Module.update_quests()
  local player = UnitName("player")

  for instance_id, quests in pairs(RAID_REQUIREMENTS) do
    local zone = GetRealZoneText(instance_id)
    local is_any = false

    for _, quest_id in pairs(quests) do
      is_any = is_any or C_QuestLog.IsQuestFlaggedCompleted(quest_id)
    end

    raid_information[player].quests[zone] = is_any
  end
end

function Module.update_status()
  local player = UnitName("player")
  local timestamp = GetServerTime()
  local count = 0

  for _, is_quest_completed in pairs(raid_information[player].quests) do
    if not is_quest_completed and character[player].level >= MIN_RAID_LEVEL then
      count = count + 1
    end
  end

  for _, instance in pairs(raid_information[player].instances) do
    if instance.timestamp - timestamp > 0 then
      count = count + 1
    end
  end

  label.text:SetFormattedText(count .. "/" .. MAX_RAID_INSTANCES)
end

function Module.enable()
  local realm = GetRealmName()
  local storage = MedusaStapleStorage.realm[realm]

  position = storage.position.raid_information
  character = storage.character
  raid_information = storage.raid_information

  label:RegisterForDrag("LeftButton")
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label:SetPoint("BOTTOMLEFT", position.x, position.y)
  label:Show()

  PROXY:RegisterEvent("MODIFIER_STATE_CHANGED")
  PROXY:RegisterEvent("QUEST_TURNED_IN")
  PROXY:RegisterEvent("UPDATE_INSTANCE_INFO")
  PROXY:RegisterEvent("BOSS_KILL")
  PROXY:SetScript("OnEvent", Module.on_event)

  Package.Observable.subscribe("position", Module.on_label_position_change)

  Module.update_instances()
  Module.update_quests()
  Module.update_status()
end

function Module.disable()
  label:Hide()

  PROXY:UnregisterEvent("MODIFIER_STATE_CHANGED")
  PROXY:UnregisterEvent("QUEST_TURNED_IN")
  PROXY:UnregisterEvent("UPDATE_INSTANCE_INFO")
  PROXY:UnregisterEvent("BOSS_KILL")
  PROXY:SetScript("OnEvent", nil)

  Package.Observable.unsubscribe("position", Module.on_label_position_change)
end

function Module.hot()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", Module.on_label_drag_start)
  label:SetScript("OnDragStop", Module.on_label_drag_stop)
  label:SetScript("OnEnter", nil)
  label:SetScript("OnLeave", nil)

  label.background:Show()
end

function Module.cold()
  GameTooltip:Hide()

  label:StopMovingOrSizing()
  label:SetScript("OnDragStart", nil)
  label:SetScript("OnDragStop", nil)
  label:SetScript("OnEnter", Module.on_label_enter)
  label:SetScript("OnLeave", Module.on_label_leave)

  label.background:Hide()
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

--------------------------------------------------------------------------------
-- Registry
--------------------------------------------------------------------------------
function Registry.MODIFIER_STATE_CHANGED(self, key, state)
  if label:IsShown() and label:IsMouseOver() and key == "LALT" then
    Module.on_label_enter(label)
  end
end

function Registry.QUEST_TURNED_IN(self, quest_id)
  if RAID_QUESTS[quest_id] then
    Module.update_quests()
    Module.update_status()
  end
end

function Registry.UPDATE_INSTANCE_INFO()
  Module.update_instances()
  Module.update_quests()
  Module.update_status()
end

function Registry.BOSS_KILL()
  RequestRaidInfo()
end

--------------------------------------------------------------------------------
-- Export
--------------------------------------------------------------------------------
Package.RaidInformation = Module